@@ -43,6 +43,7 @@ struct _GenericmenuitemPrivate {
 	GenericmenuitemCheckType   check_type;
 	GenericmenuitemState       state;
 	GenericmenuitemDisposition disposition;
+	gchar * label_text;
 };
 
 /* Private macro */
@@ -82,6 +83,12 @@ genericmenuitem_class_init (GenericmenuitemClass *klass)
 	object_class->dispose = genericmenuitem_dispose;
 	object_class->finalize = genericmenuitem_finalize;
 
+#ifdef HAVE_GTK3
+	GtkWidgetClass * widget_class = GTK_WIDGET_CLASS(klass);
+
+	gtk_widget_class_set_accessible_role(widget_class, ATK_ROLE_MENU_ITEM);
+#endif
+
 	GtkCheckMenuItemClass * check_class = GTK_CHECK_MENU_ITEM_CLASS (klass);
 
 	parent_draw_indicator = check_class->draw_indicator;
@@ -106,6 +113,14 @@ genericmenuitem_init (Genericmenuitem *self)
 	self->priv->check_type = GENERICMENUITEM_CHECK_TYPE_NONE;
 	self->priv->state = GENERICMENUITEM_STATE_UNCHECKED;
 	self->priv->disposition = GENERICMENUITEM_DISPOSITION_NORMAL;
+	self->priv->label_text = NULL;
+
+#ifndef HAVE_GTK3
+	AtkObject * aobj = gtk_widget_get_accessible(GTK_WIDGET(self));
+	if (aobj != NULL) {
+		atk_object_set_role(aobj, ATK_ROLE_MENU_ITEM);
+	}
+#endif
 
 	return;
 }
@@ -123,6 +138,8 @@ genericmenuitem_dispose (GObject *object)
 static void
 genericmenuitem_finalize (GObject *object)
 {
+	Genericmenuitem * self = GENERICMENUITEM(object);
+	g_free(self->priv->label_text);
 
 	G_OBJECT_CLASS (genericmenuitem_parent_class)->finalize (object);
 	return;
@@ -199,12 +216,43 @@ get_text_color (GenericmenuitemDisposition disposition, GtkWidget * widget)
 	return g_strdup(values[disposition].default_color);
 }
 
+/* Check to see if we've got mnemonic stuff goin' on */
+static gboolean
+has_mnemonic (const gchar * string, gboolean previous_underscore)
+{
+	if (string == NULL || string[0] == '\0') {
+		return FALSE;
+	}
+
+	if (g_utf8_get_char(string) == '_') {
+		if (previous_underscore) {
+			return has_mnemonic(g_utf8_next_char(string), FALSE);
+		} else {
+			return has_mnemonic(g_utf8_next_char(string), TRUE);
+		}
+	} else {
+		if (previous_underscore) {
+			return TRUE;
+		} else {
+			return has_mnemonic(g_utf8_next_char(string), FALSE);
+		}
+	}
+
+	return FALSE;
+}
+
 /* Set the label on the item */
 static void
 set_label (GtkMenuItem * menu_item, const gchar * in_label)
 {
 	if (in_label == NULL) return;
 
+	Genericmenuitem * item = GENERICMENUITEM(menu_item);
+	if (in_label != item->priv->label_text) {
+		g_free(item->priv->label_text);
+		item->priv->label_text = g_strdup(in_label);
+	}
+
 	/* Build a label that might include the colors of the disposition
 	   so that it gets rendered in the menuitem. */
 	gchar * local_label = NULL;
@@ -241,7 +289,12 @@ set_label (GtkMenuItem * menu_item, const gchar * in_label)
 			/* We need to put the child into a new box and
 			   make the box the child of the menu item.  Basically
 			   we're inserting a box in the middle. */
+			#ifdef HAVE_GTK3
+			GtkWidget * hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL,
+			                               get_toggle_space(GTK_WIDGET(menu_item)));
+			#else
 			GtkWidget * hbox = gtk_hbox_new(FALSE, get_toggle_space(GTK_WIDGET(menu_item)));
+			#endif
 			g_object_ref(child);
 			gtk_container_remove(GTK_CONTAINER(menu_item), child);
 			gtk_box_pack_start(GTK_BOX(hbox), child, FALSE, FALSE, 0);
@@ -259,11 +312,15 @@ set_label (GtkMenuItem * menu_item, const gchar * in_label)
 	if (labelw == NULL) {
 		/* Build it */
 		labelw = GTK_LABEL(gtk_accel_label_new(local_label));
-		gtk_label_set_use_underline(GTK_LABEL(labelw), TRUE);
 		gtk_label_set_use_markup(GTK_LABEL(labelw), TRUE);
 		gtk_misc_set_alignment(GTK_MISC(labelw), 0.0, 0.5);
-		gtk_accel_label_set_accel_widget(GTK_ACCEL_LABEL(labelw), GTK_WIDGET(menu_item));
-		gtk_label_set_markup_with_mnemonic(labelw, local_label);
+
+		if (has_mnemonic(in_label, FALSE)) {
+			gtk_label_set_use_underline(GTK_LABEL(labelw), TRUE);
+			gtk_accel_label_set_accel_widget(GTK_ACCEL_LABEL(labelw), GTK_WIDGET(menu_item));
+			gtk_label_set_markup_with_mnemonic(labelw, local_label);
+		}
+
 		gtk_widget_show(GTK_WIDGET(labelw));
 
 		/* Check to see if it needs to be in the bin for this
@@ -281,7 +338,13 @@ set_label (GtkMenuItem * menu_item, const gchar * in_label)
 			   getting in. */
 			suppress_update = TRUE;
 		} else {
-			gtk_label_set_markup_with_mnemonic(labelw, local_label);
+			if (has_mnemonic(in_label, FALSE)) {
+				gtk_label_set_use_underline(GTK_LABEL(labelw), TRUE);
+				gtk_accel_label_set_accel_widget(GTK_ACCEL_LABEL(labelw), GTK_WIDGET(menu_item));
+				gtk_label_set_markup_with_mnemonic(labelw, local_label);
+			} else {
+				gtk_label_set_markup(labelw, local_label);
+			}
 		}
 	}
 
@@ -303,25 +366,9 @@ set_label (GtkMenuItem * menu_item, const gchar * in_label)
 static const gchar *
 get_label (GtkMenuItem * menu_item)
 {
-	GtkWidget * child = gtk_bin_get_child(GTK_BIN(menu_item));
-	GtkLabel * labelw = NULL;
+	Genericmenuitem * item = GENERICMENUITEM(menu_item);
 
-	/* Try to find if we have a label already */
-	if (child != NULL) {
-		if (GTK_IS_LABEL(child)) {
-			/* We've got a label, let's update it. */
-			labelw = GTK_LABEL(child);
-		} else if (GTK_IS_BOX(child)) {
-			/* Look for the label in the box */
-			gtk_container_foreach(GTK_CONTAINER(child), set_label_helper, &labelw);
-		}
-	}
-
-	if (labelw != NULL) {
-		return gtk_label_get_label(labelw);
-	}
-
-	return NULL;
+	return item->priv->label_text;
 }
 
 /* Make sure we don't toggle when there is an
@@ -348,18 +395,29 @@ genericmenuitem_set_check_type (Genericmenuitem * item, GenericmenuitemCheckType
 	}
 
 	item->priv->check_type = check_type;
+	AtkObject * aobj = gtk_widget_get_accessible(GTK_WIDGET(item));
 
 	switch (item->priv->check_type) {
 	case GENERICMENUITEM_CHECK_TYPE_NONE:
 		/* We don't need to do anything here as we're queuing the
 		   draw and then when it draws it'll avoid drawing the
 		   check on the item. */
+
+		if (aobj != NULL) {
+			atk_object_set_role(aobj, ATK_ROLE_MENU_ITEM);
+		}
 		break;
 	case GENERICMENUITEM_CHECK_TYPE_CHECKBOX:
 		gtk_check_menu_item_set_draw_as_radio(GTK_CHECK_MENU_ITEM(item), FALSE);
+		if (aobj != NULL) {
+			atk_object_set_role(aobj, ATK_ROLE_CHECK_MENU_ITEM);
+		}
 		break;
 	case GENERICMENUITEM_CHECK_TYPE_RADIO:
 		gtk_check_menu_item_set_draw_as_radio(GTK_CHECK_MENU_ITEM(item), TRUE);
+		if (aobj != NULL) {
+			atk_object_set_role(aobj, ATK_ROLE_RADIO_MENU_ITEM);
+		}
 		break;
 	default:
 		g_warning("Generic Menuitem invalid check type: %d", check_type);
@@ -457,7 +515,12 @@ genericmenuitem_set_image (Genericmenuitem * menu_item, GtkWidget * image)
 			/* We need to put the child into a new box and
 			   make the box the child of the menu item.  Basically
 			   we're inserting a box in the middle. */
+			#ifdef HAVE_GTK3
+			GtkWidget * hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL,
+			                               get_toggle_space(GTK_WIDGET(menu_item)));
+			#else
 			GtkWidget * hbox = gtk_hbox_new(FALSE, get_toggle_space(GTK_WIDGET(menu_item)));
+			#endif
 			g_object_ref(child);
 			gtk_container_remove(GTK_CONTAINER(menu_item), child);
 			gtk_box_pack_end(GTK_BOX(hbox), child, TRUE, TRUE, 0);
